% W-H system simulation program
% nonlinear system y = av1*x + av2*x^2 + av3*x^3 + av4*x^4+ av5*x^5,
%
% x - L - x1 - NL - x2 - R -y
%
clear all; clf; rand('seed',0);
%
av=[1 .1*[0.2 0.5 0.05 0.05]];
% 'irg' - ideal gaussian, 'rms' - multisine, 'rmso1' - odd, 'rmso2' - odd-odd, 

ML=100;                   %number of measurements
npoints=1024*2*4; 		% number of time domain points
n=npoints/2;     		% maximum number of frequencies
nfreq=fix(npoints/10);
junit=sqrt(-1);
%
% frequency axes
fv=[1:nfreq]';
fvo1=[1:2:nfreq]'; 
fvo2=[1:4:nfreq]';
fveven=[2:2:nfreq]';
fvodd=[3:4:nfreq]';
    % fvo2=[1:64:nfreq]'; Nf=length(fvo2);     % rare RMS for colored RMS, Hammerstein
%
fvrel=0.1*fv/nfreq; zet=exp(-sqrt(-1)*2*pi*fvrel);
%
% W-H system design, 'L' input dynamics, 'R' output dynamics
%[b1,a1]=butter(3,0.03,'high'); 
%[b2,a2]=butter(3,0.07);
%
[b1,a1]=butter(3,[0.03 0.07]); % on 0 ... 0.1 axis 3dB points are 0.015 0.035
%[b1,a1]=cheby1(7,1.3,0.07);
[b1,a1]=butter(0,0.1);
%[b2,a2]=ellip(3,2,20,0.07);
[b2,a2]=butter(3,[0.03 0.07]);
%
Lb1=length(b1); HLbf=0; for i=1:Lb1, HLbf=HLbf+b1(i)*zet.^(i-1); end;
La1=length(a1); HLaf=0; for i=1:La1, HLaf=HLaf+a1(i)*zet.^(i-1); end;
HLf=HLbf./HLaf;
%
Lb2=length(b2); HRbf=0; for i=1:Lb2, HRbf=HRbf+b2(i)*zet.^(i-1); end;
La2=length(a2); HRaf=0; for i=1:La2, HRaf=HRaf+a2(i)*zet.^(i-1); end; 
HRf=HRbf./HRaf;
%
LRf=HLf.*HRf;
%
% related dynamic systems
RDS1=1;
RDS3=3*sum(abs(HLf(fv)).^2)/length(fv);
RDS5=5*3*(sum(abs(HLf(fv)).^2)/length(fv))^2;
RDStr=av(1)*RDS1+av(3)*RDS3+av(5)*RDS5;

RDS1o1=1;
RDS3o1=3*sum(abs(HLf(fvo1)).^2)/length(fvo1);
RDS5o1=5*3*(sum(abs(HLf(fvo1)).^2)/length(fvo1))^2;
RDStro1=av(1)*RDS1o1+av(3)*RDS3o1+av(5)*RDS5o1;

RDS1o2=1;
RDS3o2=3*sum(abs(HLf(fvo2)).^2)/length(fvo2);
RDS5o2=5*3*(sum(abs(HLf(fvo2)).^2)/length(fvo2))^2;
RDStro2=av(1)*RDS1o2+av(3)*RDS3o2+av(5)*RDS5o2;
%
figure(1),
 subplot(2,2,1), plot(fvrel,db(HLf)); grid; axis([0 .1 -50 10]);
 subplot(2,2,2), plot(fvrel,db(HRf)); grid; axis([0 .1 -50 10]);
 subplot(2,2,4), plot(fvrel,db(LRf)); grid; axis([0 .1 -50 10]);
 title('L(f), R(f), LR(f)');
 pause;
%
Xrms=zeros(npoints,1); xrms=Xrms;
Xrmso1=zeros(npoints,1); xrmso1=Xrms;
Xrmso2=zeros(npoints,1); xrmso2=Xrms;
Xirg=zeros(npoints,1); xirg=Xrms;
%
Grms=zeros(length(fv),1); 
Grmso1=zeros(length(fvo1),1); 
Grmso2=zeros(length(fvo2),1); 
Girg=zeros(length(fv),1); 
Syxirg=Girg; Sxxirg=Girg; 
GrmsAv=0*Grms; Grmso1Av=0*Grmso1; Grmso2Av=0*Grmso2; GirgAv=0*Girg; 
Veveno1=zeros(length(fveven),1);
Voddo2=zeros(length(fvodd),1);

VG=zeros(length(fv),1);
VGo2=zeros(length(fvo2),1);
%
for ll=1:ML, ll
% xrms, Xrms - random phase multisine
 fi=rand(length(fv),1)*2*pi; Zx=exp(-junit*fi); Xrms=0*Xrms;
 Xrms(fv+1)=Zx; Xrms(2*n+1-fv)=conj(Xrms(1+fv));
 xrms=real(ifft(Xrms)); Ax=sqrt(sum(xrms.^2)/npoints);  xrms=xrms/Ax;
 Xrms=fft(xrms); AW=abs(Xrms)/max(abs(Xrms));

% xrmso1, Xrmso1 - random phase odd freq multisine
 fio1=rand(length(fvo1),1)*2*pi; Zxo1=exp(-junit*fio1);
 Xrmso1=0*Xrmso1;
 Xrmso1(fvo1+1)=Zxo1; Xrmso1(2*n+1-fvo1)=conj(Xrmso1(1+fvo1));
 xrmso1=real(ifft(Xrmso1)); Ax=sqrt(sum(xrmso1.^2)/npoints); xrmso1=xrmso1/Ax;
 Xrmso1=fft(xrmso1); 

% xrmso2, Xrmso2 - random phase odd-odd freq multisine
 fio2=rand(length(fvo2),1)*2*pi; 
    %ampve=Nf*ones(size(fvo2))./fvo2;  
    %Zxo=ampve.*exp(-junit*fio2); % coloring
                                  % Hammerstein biased, when
                                  % colored, bec. non separable
 Zxo2=exp(-junit*fio2);
 Xrmso2=0*Xrmso2;
 Xrmso2(fvo2+1)=Zxo2; Xrmso2(2*n+1-fvo2)=conj(Xrmso2(1+fvo2));
 xrmso2=real(ifft(Xrmso2)); Ax=sqrt(sum(xrmso2.^2)/npoints); xrmso2=xrmso2/Ax;
 Xrmso2=fft(xrmso2); 

% xirg, Xirg - ideal filtered Gaussian signal
 xaux=randn(length(xrms),1); Xaux=fft(xaux); Xirg=Xaux.*AW;
 xirg=real(ifft(Xirg)); Ax=sqrt(sum(xirg.^2)/npoints); xirg=xirg/Ax;
 Xirg=fft(xirg);

%figure(3),
%subplot(2,2,2), plot(db(Xrms)); xlabel('Xrms','Fontsize',14);
%title('Xrms');
%subplot(2,2,3), plot(db(Xrmso1)); xlabel('Xrmso1','Fontsize',14);
%title('Xrmso1');
%subplot(2,2,4), plot(db(Xrmso2)); xlabel('Xrmso2','Fontsize',14);
%title('Xrmso2');
%subplot(2,2,1), plot(db(Xirg)); xlabel('Xirg','Fontsize',14);
%title('Xirg');
%drawnow;
%pause;
%
%figure(4),
%subplot(2,2,2), plot(xrms); xlabel('xrms','Fontsize',14); grid;
%title('Xrms');
%subplot(2,2,3), plot(xrmso1); xlabel('xrmso1','Fontsize',14); grid;
%title('Xrmso1');
%subplot(2,2,4), plot(xrmso2); xlabel('xrmso2','Fontsize',14); grid;
%title('Xrmso2');
%subplot(2,2,1), plot(xirg); xlabel('xirg','Fontsize',14); grid;
%title('Xirg');
%drawnow;
%pause;
%
% passing signals through input dynamics

xrmsfull=[xrms; xrms; xrms; xrms];              x1rms=filter(b1,a1,xrmsfull);
xrmsfull=[xrmso1; xrmso1; xrmso1; xrmso1];      x1rmso1=filter(b1,a1,xrmsfull);
xrmsfull=[xrmso2; xrmso2; xrmso2; xrmso2];      x1rmso2=filter(b1,a1,xrmsfull);
xrmsfull=[xirg; xirg; xirg; xirg];              x1irg=filter(b1,a1,xrmsfull);

Nfull=length(xrmsfull);
Nfullminrec=Nfull-npoints+1;

% passing signals through nonlinear system
x2rms=0*x1rms; x2rmso1=0*x1rms; x2irg=0*x1rms; x2rmso2=0*x1rms; 

if av(1)~=0,
  x2rms=x2rms+av(1)*x1rms; x2rmso1=x2rmso1+av(1)*x1rmso1;
  x2irg=x2irg+av(1)*x1irg; x2rmso2=x2rmso2+av(1)*x1rmso2;
end;
if av(2)~=0,
  x2rms=x2rms+av(2)*x1rms.^2; x2rmso1=x2rmso1+av(2)*x1rmso1.^2;
  x2irg=x2irg+av(2)*x1irg.^2;  x2rmso2=x2rmso2+av(2)*x1rmso2.^2;
end;
if av(3)~=0,
  x2rms=x2rms+av(3)*x1rms.^3; x2rmso1=x2rmso1+av(3)*x1rmso1.^3;
  x2irg=x2irg+av(3)*x1irg.^3; x2rmso2=x2rmso2+av(3)*x1rmso2.^3;
end;
if av(4)~=0,
  x2rms=x2rms+av(4)*x1rms.^4; x2rmso1=x2rmso1+av(4)*x1rmso1.^4;
  x2irg=x2irg+av(4)*x1irg.^4; x2rmso2=x2rmso2+av(4)*x1rmso2.^4;
end;
if av(5)~=0,
  x2rms=x2rms+av(5)*x1rms.^5; x2rmso1=x2rmso1+av(5)*x1rmso1.^5;
  x2irg=x2irg+av(5)*x1irg.^5; x2rmso2=x2rmso2+av(5)*x1rmso2.^5;
end;
%
% passing signals through output dynamics
%
yrms=filter(b2,a2,x2rms);       Yrms=fft(yrms(Nfullminrec:Nfull));
yrmso1=filter(b2,a2,x2rmso1);   Yrmso1=fft(yrmso1(Nfullminrec:Nfull));
yrmso2=filter(b2,a2,x2rmso2);   Yrmso2=fft(yrmso2(Nfullminrec:Nfull));
yirg=filter(b2,a2,x2irg);       Yirg=fft(yirg(Nfullminrec:Nfull));
%
%figure(5),
%subplot(2,2,2), plot(yrms(Nfullminrec:Nfull)); xlabel('yrms','Fontsize',14); grid;
%title('yrms');
%subplot(2,2,3), plot(yrmso1(Nfullminrec:Nfull)); xlabel('yrmso1','Fontsize',14); grid;
%title('yrmso1');
%subplot(2,2,4), plot(yrmso2(Nfullminrec:Nfull)); xlabel('yrmso2','Fontsize',14); grid;
%title('yrmso2');
%subplot(2,2,1), plot(yirg(Nfullminrec:Nfull)); xlabel('yirg','Fontsize',14); grid;
%title('yirg');
%drawnow;
%pause;

%figure(6),
%subplot(2,2,2), plot(db(Yrms)); xlabel('Yrms','Fontsize',14); grid;
%title('Yrms');
%subplot(2,2,3), plot(db(Yrmso1)); xlabel('Yrmso1','Fontsize',14); grid;
%title('Yrmso1');
%subplot(2,2,4), plot(db(Yrmso2)); xlabel('Yrmso2','Fontsize',14); grid;
%title('Yrmso2');
%subplot(2,2,1), plot(db(Yirg)); xlabel('Yirg','Fontsize',14); grid;
%title('Yirg');
%drawnow;
%pause;

X1rms=fft(x1rms(1:npoints));
X2rms=fft(x2rms(1:npoints));

%figure(7),
%subplot(2,2,2), plot(db(X1rms)); xlabel('X1rms','Fontsize',14); grid;
%title('X1rms');
%subplot(2,2,3), plot(db(X2rms)); xlabel('X2rms','Fontsize',14); grid;
%title('X2rms');
%subplot(2,2,4), plot(db(Yrms)); xlabel('Yrms','Fontsize',14); grid;
%title('Yrms');
%subplot(2,2,1), plot(db(Xrms)); xlabel('Xrms','Fontsize',14); grid;
%title('Xrms');
%drawnow;
%pause;

% computing estimates of the system transfer characteristics
 Grms=Yrms(fv+1)./Xrms(fv+1);
 Grmso1=Yrmso1(fvo1+1)./Xrmso1(fvo1+1);
 Grmso2=Yrmso2(fvo2+1)./Xrmso2(fvo2+1);
 
 GrmsAv=GrmsAv+(Grms-GrmsAv)/ll;
 Grmso1Av=Grmso1Av+(Grmso1-Grmso1Av)/ll;
 Grmso2Av=Grmso2Av+(Grmso2-Grmso2Av)/ll;
 
 Veveno1=Veveno1+(abs(Yrmso1(fveven+1)).^2-Veveno1)/ll;
 Voddo2=Voddo2+(abs(Yrmso2(fvodd+1)).^2-Voddo2)/ll;
 
 VG=VG+((abs(Grms-GrmsAv)).^2-VG)/ll;
 VGo2=VGo2+((abs(Grmso2-Grmso2Av)).^2-VGo2)/ll;

% for gaussian signals
 Syxirg=Yirg(fv+1).*conj(Xirg(fv+1));
 Sxxirg=Xirg(fv+1).*conj(Xirg(fv+1));
 Girg=Syxirg./Sxxirg;
 
 GirgAv=GirgAv+(Girg-GirgAv)/ll;

 %figure(10),
 %plot(fv,db(RDStr*LRf),'g',fv,db(Girg(:)),'k',fv,db(Grms),'b',fvo1,db(Grmso1(:)),'c',...
 %    fvo2,db(Grmso2(:)),'r',fveven,db10(Veveno1(:)/npoints),'m',...
 %    fvodd,db10(Voddo2(:)/npoints),'k',fvo2,db10(VGo2),'b',fv,db10(VG),'r');
 %axis([0 length(fv) -30 10]);
 %grid; title('b - rms, y - rmso1, g - irg, c - cfrg, '); 
 %title('RDStr*LRf  Girg  Grms  Grmso1 Grmso2  Veveno1  Voddo2  VGo2  VG');
 %drawnow;
%pause;
%
strcol='c';
 figure(11),
 subplot(1,4,2), plot(fv,db(Grms(:)),strcol,fv,db(RDStr*LRf),'r',fv,db(GrmsAv(:)),'b');
 axis([0 1000 -30 20]); grid; title('RMS','Fontsize',14);
 subplot(1,4,1), plot(fv,db(Girg(:)),strcol,fv,db(RDStr*LRf),'r',fv,db(GirgAv(:)),'b');
 axis([0 1000 -30 20]); grid; title('Gauss','Fontsize',14);
 subplot(1,4,3), plot(fvo1,db(Grmso1(:)),strcol,fv,db(RDStro1*LRf),'r',fvo1,db(Grmso1Av(:)),'b');
 axis([0 1000 -30 20]); grid; title('odd RMS','Fontsize',14);
 subplot(1,4,4), plot(fvo2,db(Grmso2(:)),strcol,fv,db(RDStro2*LRf),'r',fvo2,db(Grmso2Av(:)),'b');
 axis([0 1000 -30 20]); grid; title('odd-odd RMS','Fontsize',14);
 drawnow;
%pause;
end;

figure(12),
subplot(1,4,2), plot(fv,db(RDStr*LRf),'r',fv,db(GrmsAv(:)),'b');
axis([0 1000 -30 20]); grid; title('RMS','Fontsize',14);
subplot(1,4,1), plot(fv,db(RDStr*LRf),'r',fv,db(GirgAv(:)),'b');
axis([0 1000 -30 20]); grid; title('Gauss','Fontsize',14);
subplot(1,4,3), plot(fv,db(RDStro1*LRf),'r',fvo1,db(Grmso1Av(:)),'b');
axis([0 1000 -30 20]); grid; title('odd RMS','Fontsize',14);
subplot(1,4,4), plot(fv,db(RDStro2*LRf),'r',fvo2,db(Grmso2Av(:)),'b');
axis([0 1000 -30 20]); grid; title('odd-odd RMS','Fontsize',14);
%pause;

%figure(13),
%plot(fv,db(LRf),'g',fv,db(RDStr*LRf),'r',fv,db(GrmsAv(:)),'b');
%axis([0 1000 -30 10]); grid; title('RMS','Fontsize',14);

%figure(14), 
%subplot(2,1,1), plot(xrms); axis([0 length(xrms) -10 10]);
%grid; title('Input RMS','Fontsize',14);
%subplot(2,1,2), plot(yrms); axis([0 length(yrms) -10 10]);
%grid; title('Output RMS','Fontsize',14);
%pause;

%figure(15), 
%subplot(2,1,1), plot(db(Xrms));
%grid; title('Input RMS','Fontsize',14);
%subplot(2,1,2), plot(db(Yrms));
%grid; title('Output RMS','Fontsize',14);
%pause;
%
xv=[-3:.01:3];
yv=av(1)*xv+av(2)*xv.^2+av(3)*xv.^3+av(4)*xv.^4+av(5)*xv.^5;
figure(16), plot(xv,yv,'o'); grid;
title('Nonlinearity','Fontsize',14);
%pause;

% % power in NL measured around static NL
AA1=sqrt(sum(x1rms.^2)/npoints); AA2=sqrt(sum(x2rms.^2)/npoints); 100*(AA2-AA1)/AA1

